//
// Program that implements Akelarre block cipher
//
// Block size is 128 bits (four 32-bit words)
//
// Note that encryption and decrytion subkeys are very different
// Also note that there are some inconsistencies in the 
// published design(s) of the cipher
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// maximum number of rounds (typically, use 4 rounds)
#define MAX_ROUNDS 10
// max number of subkeys (depends on number of rounds)
#define MAX_SUBKEYS (13*MAX_ROUNDS + 9)
// print input
#define PR_INPUT
// print output
#define PR_OUTPUT
// print rotations
#define PR_ROTATION
// print encryption and decryption subkeys
#define PR_SUBKEYS
// print number of rounds
#define PR_ROUNDS


void akelarre(unsigned long int inputBlock[], 
			  unsigned long int outputBlock[], 
			  unsigned long int subkey[],
			  int rounds);
			  
void AR(unsigned long int w[],
		unsigned long int subkey[], 
		int i);
			  
void rotate(unsigned long int x[], 
			unsigned long int key);

void getSubkeys(unsigned long int key[], 
				unsigned long int encSubkey[], 
				unsigned long int decSubkey[], 
				int rounds);

unsigned long int neg(unsigned long int x);


int main(int argc, const char *argv[])
{
	FILE *in,
		 *out;

    int a,
		i,
        j,
		k,
		flag,
		rounds,
		direction;

	unsigned long int inputBlock[4],
					  outputBlock[4],
					  encSubkey[MAX_SUBKEYS],
					  decSubkey[MAX_SUBKEYS];

	unsigned long int plain[4] = {0x71706f6e,0x6c6b6a69,0x68676665,0x7a797877},
//					  plain[4] = {0x61616262,0x68696362,0x61606e6f,0x71717070},
//					  plain[4] = {0x61626364,0x71706f6e,0x6d6b6a6c,0x6771646a},
					  cipher[4],
					  decrypt[4];

	// internal representation of the key
	unsigned long int key[4] = {0x76543210,0xdeadbeef,0xfedcba98,0xa55aa55a};

	char infname[128],
		 outfname[128];
		 
	// x[] must be "unsigned char", not just "char"
	unsigned char x[4];

    if(argc != 5)
    {
oops:   fprintf(stderr, "\nUsage: %s rounds infile outfile direction\n\n", argv[0]);
        fprintf(stderr, "where rounds == number of rounds\n");
        fprintf(stderr, "      infile == name of input file\n");
        fprintf(stderr, "      outfile == name of output file\n");
        fprintf(stderr, "      direction == 0 to encrypt, 1 to decrypt\n\n");
        fprintf(stderr, "It is recommended to use 4 rounds\n\n");
        exit(0);
    }

	rounds = atoi(argv[1]);
	if(rounds < 0 || rounds > MAX_ROUNDS)
	{
		fprintf(stderr, "\n\nError! Number of rounds must be between 0 and %d\n\n",
			MAX_ROUNDS);
		goto oops;
	
	}// end if
	
#ifdef PR_ROUNDS
	//printf("rounds = %d\n", rounds);
#endif

	
	direction = atoi(argv[4]);
	
	if(direction < 0 || direction > 1)
	{
		fprintf(stderr, "\n\nError! The value of direction must be either 0 or 1, ");
		fprintf(stderr, "value read was %d\n\n", direction);
	
		goto oops;
	
	}// end if
	
	sprintf(infname,"%s", argv[2]);
	
    
    in = fopen(infname, "r");
    if(in == NULL)
    {
        fprintf(stderr, "\n\nError opening file %s\nTry again\n\n", infname);
        goto oops;
    }

	sprintf(outfname,"%s", argv[3]);

    out = fopen(outfname, "w");
    if(out == NULL)
    {
        fprintf(stderr, "\n\nError opening file %s\nTry again\n\n", outfname);
        goto oops;
    }

	// get both encryption and decryption subkeys
	getSubkeys(key, encSubkey, decSubkey, rounds);

	flag = 1;
    while(flag == 1)
    {
		for(j = 0; j < 4; ++j)
		{
			inputBlock[j] = 0;
		}
		for(j = 0; j < 4; ++j)
		{
			x[0] = x[1] = x[2] = x[3] = 0;
			a = fscanf(in, "%c%c%c%c", &x[0], &x[1], &x[2], &x[3]);
//			printf("input = %02x %02x %02x %02x\n", x[0], x[1], x[2], x[3]);
			if(a > 0)
			{
				inputBlock[j] = ((unsigned long int)x[0] << 24L)
								| ((unsigned long int)x[1] << 16L) 
								| ((unsigned long int)x[2] << 8L) 
								| (unsigned long int)x[3];
			}
			else
			{
				break;
			}
		}
		if(j < 4)
		{
			if(j == 0) break;

			flag = 0;
		}

#ifdef PR_INPUT
		printf("input  = ");
		for(i = 0; i < 4; ++i)
		{
			printf("%08lx ", inputBlock[i]);
		}
		printf("\n");
#endif
		
		if(direction == 0)// encrypt
		{
			akelarre(inputBlock, outputBlock, encSubkey, rounds);
		}
		else// decrypt
		{
			akelarre(inputBlock, outputBlock, decSubkey, rounds);
		}

#ifdef PR_OUTPUT		
		printf("output = ");
		for(i = 0; i < 4; ++i)
		{
			printf("%08lx ", outputBlock[i]); //08x
		}
		printf("\n");
#endif

		for(j = 0; j < 4; ++j)
		{
			fprintf(out, "%c", (unsigned char)((outputBlock[j] >> 24L) & 0xff));
			fprintf(out, "%c", (unsigned char)((outputBlock[j] >> 16L) & 0xff));
			fprintf(out, "%c", (unsigned char)((outputBlock[j] >>  8L) & 0xff));
			fprintf(out, "%c", (unsigned char)((outputBlock[j] >>  0L) & 0xff));
		}

	}// end while
	
	fclose(in);
	fclose(out);

}// end main


void akelarre(unsigned long int inputBlock[], 
			  unsigned long int outputBlock[], 
			  unsigned long int subkey[],
			  int rounds)
{
	int i;
	unsigned long int x[4],
					  w[2];

	// input transformation
	x[0] = inputBlock[0] + subkey[0];
	x[1] = inputBlock[1] ^ subkey[1];
	x[2] = inputBlock[2] ^ subkey[2];
	x[3] = inputBlock[3] + subkey[3];

	// rounds: 0 thru rounds - 1
	for(i = 0; i < rounds; ++i)
	{
		rotate(x, subkey[13*i + 4]);
		// inputs to AR
		w[0] = x[0] ^ x[2];
		w[1] = x[1] ^ x[3];
		
		AR(w, subkey, i);
		
		x[0] ^= w[1];
		x[1] ^= w[0];
		x[2] ^= w[1];
		x[3] ^= w[0];
	}
	
	// output transformation
	rotate(x, subkey[13*rounds + 4]);

	outputBlock[0] = x[0] + subkey[13*rounds + 5];
	outputBlock[1] = x[1] ^ subkey[13*rounds + 6];
	outputBlock[2] = x[2] ^ subkey[13*rounds + 7];
	outputBlock[3] = x[3] + subkey[13*rounds + 8];

}// end akelarre


void AR(unsigned long int w[],
		unsigned long int subkey[], 
		int i)
{
	int t;
	
	// new w[1]
	t = w[0] & 0x1f;
	if(t > 0)
	{
		w[1] = ((w[1] & 0xfffffffe) << t) | (w[1] >> (31 - t)) | (w[1] & 0x1);
	}
	w[1] += subkey[13*i + 5];
	t = (w[0] >> 5) & 0x1f;
	if(t > 0)
	{
		w[1] = (w[1] << t) | ((w[1] & 0x7fffffff) >> (31 - t)) | (w[1] & 0x80000000);
	}
	w[1] += subkey[13*i + 6];

	t = (w[0] >> 10) & 0x1f;
	if(t > 0)
	{
		w[1] = ((w[1] & 0xfffffffe) << t) | (w[1] >> (31 - t)) | (w[1] & 0x1);
	}
	w[1] += subkey[13*i + 7];
	t = (w[0] >> 15) & 0x1f;
	if(t > 0)
	{
		w[1] = (w[1] << t) | ((w[1] & 0x7fffffff) >> (31 - t)) | (w[1] & 0x80000000);
	}
	w[1] += subkey[13*i + 8];

	t = (w[0] >> 20) & 0xf;
	if(t > 0)
	{
		w[1] = ((w[1] & 0xfffffffe) << t) | (w[1] >> (31 - t)) | (w[1] & 0x1);
	}
	w[1] += subkey[13*i + 9];
	t = (w[0] >> 24) & 0xf;
	if(t > 0)
	{
		w[1] = (w[1] << t) | ((w[1] & 0x7fffffff) >> (31 - t)) | (w[1] & 0x80000000);
	}
	w[1] += subkey[13*i + 10];

	t = (w[0] >> 28) & 0xf;
	if(t > 0)
	{
		w[1] = ((w[1] & 0xfffffffe) << t) | (w[1] >> (31 - t)) | (w[1] & 0x1);
	}

	// new w[0]
	t = w[1] & 0x1f;
	if(t > 0)
	{
		w[0] = ((w[0] & 0xfffffffe) << t) | (w[0] >> (31 - t)) | (w[0] & 0x1);
	}
	w[0] += subkey[13*i + 11];
	t = (w[1] >> 5) & 0x1f;
	if(t > 0)
	{
		w[0] = (w[0] << t) | ((w[0] & 0x7fffffff) >> (31 - t)) | (w[0] & 0x80000000);
	}
	w[0] += subkey[13*i + 12];

	t = (w[1] >> 10) & 0x1f;
	if(t > 0)
	{
		w[0] = ((w[0] & 0xfffffffe) << t) | (w[0] >> (31 - t)) | (w[0] & 0x1);
	}
	w[0] += subkey[13*i + 13];
	t = (w[1] >> 15) & 0x1f;
	if(t > 0)
	{
		w[0] = (w[0] << t) | ((w[0] & 0x7fffffff) >> (31 - t)) | (w[0] & 0x80000000);
	}
	w[0] += subkey[13*i + 14];

	t = (w[1] >> 20) & 0xf;
	if(t > 0)
	{
		w[0] = ((w[0] & 0xfffffffe) << t) | (w[0] >> (31 - t)) | (w[0] & 0x1);
	}
	w[0] += subkey[13*i + 15];
	t = (w[1] >> 24) & 0xf;
	if(t > 0)
	{
		w[0] = (w[0] << t) | ((w[0] & 0x7fffffff) >> (31 - t)) | (w[0] & 0x80000000);
	}
	w[0] += subkey[13*i + 16];

	t = (w[1] >> 28) & 0xf;
	if(t > 0)
	{
		w[0] = ((w[0] & 0xfffffffe) << t) | (w[0] >> (31 - t)) | (w[0] & 0x1);
	}

}// end AR


void rotate(unsigned long int x[], unsigned long int key)
{
	int t;

	unsigned long int y[4];

	t = key & 0x7f;

	if(t <= 32)// 0 <= t <= 32
	{
		y[0] = (x[0] << t) | (x[1] >> (32 - t));
		y[1] = (x[1] << t) | (x[2] >> (32 - t));
		y[2] = (x[2] << t) | (x[3] >> (32 - t));
		y[3] = (x[3] << t) | (x[0] >> (32 - t));
	}
	else// 33 <= t <= 127
	{
		if(t <= 64)// 33 <= t <= 64
		{
			y[0] = (x[1] << (t - 32)) | (x[2] >> (64 - t));
			y[1] = (x[2] << (t - 32)) | (x[3] >> (64 - t));
			y[2] = (x[3] << (t - 32)) | (x[0] >> (64 - t));
			y[3] = (x[0] << (t - 32)) | (x[1] >> (64 - t));	
		}
		else// 65 <= t <= 127
		{
			if(t <= 96)// 65 <= t <= 96
			{
				y[0] = (x[2] << (t - 64)) | (x[3] >> (96 - t));
				y[1] = (x[3] << (t - 64)) | (x[0] >> (96 - t));
				y[2] = (x[0] << (t - 64)) | (x[1] >> (96 - t));
				y[3] = (x[1] << (t - 64)) | (x[2] >> (96 - t));		
			}
			else// 97 <= t <= 127
			{
				y[0] = (x[3] << (t - 96)) | (x[0] >> (128 - t));
				y[1] = (x[0] << (t - 96)) | (x[1] >> (128 - t));
				y[2] = (x[1] << (t - 96)) | (x[2] >> (128 - t));
				y[3] = (x[2] << (t - 96)) | (x[3] >> (128 - t));			
			}
		}
	}
	
#ifdef PR_ROTATION
	printf("t = rotation = %d\n", t);
	printf("x = %08lx %08lx %08lx %08lx\n", x[0], x[1], x[2], x[3]);
	printf("y = %08lx %08lx %08lx %08lx\n", y[0], y[1], y[2], y[3]);
#endif
	
	x[0] = y[0];
	x[1] = y[1];
	x[2] = y[2];
	x[3] = y[3];

}// end rotate


void getSubkeys(unsigned long int key[], 
				unsigned long int encSubkey[], 
				unsigned long int decSubkey[], 
				int rounds)
{
	int i,
		j,
		cnt;
		
	unsigned long int t0,
					  t1;
		
	unsigned long int u[8],
					  v[8],
					  a[2];

	// constants
	a[0] = 0xa49ed284;
	a[1] = 0x735203de;

	// assuming key is 128 bits
	for(i = 0; i < 4; ++i)
	{
		u[2*i] = v[2*i] = (key[i] & 0xffff0000) >> 8;
		u[2*i + 1] = v[2*i + 1] = (key[i] & 0xffff) << 8;
	}
	cnt = 0;
	while(cnt < (13*rounds + 9))
	{
		for(i = 0; i < 8; ++i)
		{
			u[i] = ((u[i] >> 8) & 0xffff)*((u[i] >> 8) & 0xffff) + a[0];
			v[i] = ((v[i] >> 8) & 0xffff)*((v[i] >> 8) & 0xffff) + a[1];
		}
		for(i = 0; i < 8; ++i)
		{
			t0 = u[i];
			t1 = v[(i + 1) & 0x7];
			encSubkey[cnt] = ((t0 & 0xff) << 24) 
								| ((t0 & 0xff000000) >> 8) 
								| ((t1 & 0xff) << 8)
								| ((t1 & 0xff000000) >> 24);
			++cnt;
			if(cnt >= (13*rounds + 9))
			{
				break;
			}

		}// next i

	}// end while
	
	// input transformation
	decSubkey[0] = -encSubkey[13*rounds + 5];
	decSubkey[1] = encSubkey[13*rounds + 6];
	decSubkey[2] = encSubkey[13*rounds + 7];
	decSubkey[3] = -encSubkey[13*rounds + 8];
	
	// round i
	for(i = 0; i < rounds; ++i)
	{
		decSubkey[4 + 13*i] = neg(encSubkey[13*(rounds - i) + 4]);
		for(j = 1; j < 13; ++j)
		{
			decSubkey[4 + 13*i + j] = encSubkey[13*(rounds - i - 1) + 4 + j];
		}
	
	}// next i
	
	// output transformation
	decSubkey[13*rounds + 4] = neg(encSubkey[4]);
	decSubkey[13*rounds + 5] = -encSubkey[0];
	decSubkey[13*rounds + 6] = encSubkey[1];
	decSubkey[13*rounds + 7] = encSubkey[2];
	decSubkey[13*rounds + 8] = -encSubkey[3];

#ifdef PR_SUBKEYS
	printf("\n");
	for(i = 0; i < 13*rounds + 9; ++i)
	{
		printf("encSubkey[%d] = %08lx\n", i, encSubkey[i]);
	}
	printf("\n");
	for(i = 0; i < 13*rounds + 9; ++i)
	{
		printf("decSubkey[%d] = %08lx\n", i, decSubkey[i]);
	}
	printf("\n");
#endif

}// end getSubkeys


// want negative of low 7 bits
unsigned long int neg(unsigned long int x)
{
	unsigned long int y;
	int t;
	
	t = x & 0x7f;
	t = (128 - t) & 0x7f;

	y = (x & 0xffffff80) | (unsigned long int)t;

	return(y);
	
}// end inv
